package se.su.dsv.scipro.finalseminar;

import com.google.common.eventbus.EventBus;
import com.google.inject.persist.Transactional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.exceptions.InvalidFileFormatException;
import se.su.dsv.scipro.file.*;
import se.su.dsv.scipro.plagiarism.PlagiarismControl;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.util.Date;

public class FinalSeminarUploadControllerImpl implements FinalSeminarUploadController {

    private static final Logger LOGGER = LoggerFactory.getLogger(FinalSeminarUploadControllerImpl.class);

    private final FileDescriptionService fileDescriptionService;
    private final FinalSeminarService finalSeminarService;
    private final ProjectService projectService;
    private final EventBus eventBus;
    private final ProjectFileService projectFileService;
    private final PlagiarismControl plagiarismControl;
    private final FinalSeminarSettingsService finalSeminarSettingsService;

	@Inject
    public FinalSeminarUploadControllerImpl(
            ProjectService projectService,
            FileDescriptionService fileDescriptionService,
            FinalSeminarSettingsService finalSeminarSettingsService,
            FinalSeminarService finalSeminarService,
            EventBus eventBus, ProjectFileService projectFileService,
            final PlagiarismControl plagiarismControl)
    {
        this.projectService = projectService;
        this.fileDescriptionService = fileDescriptionService;
        this.finalSeminarSettingsService = finalSeminarSettingsService;
        this.finalSeminarService = finalSeminarService;
        this.eventBus = eventBus;
        this.projectFileService = projectFileService;
        this.plagiarismControl = plagiarismControl;
    }

    @Override
    public FileDescription storeSeminarDocument(FileUpload fileUpload, FinalSeminar seminar) {
        assertParticipant(fileUpload.getUploader(), seminar);
        assertMimeType(fileUpload.getContentType());
        seminar = finalSeminarService.findOne(seminar.getId());
        if (seminar.getDocument() != null){
            throw new IllegalStateException("A file has already been uploaded");
        }

        FileDescription fd = fileDescriptionService.storeFile(fileUpload);
        setSeminarDocument(fd, seminar);
        projectFileService.promote(fd, seminar.getProject(), FileSource.FINAL_SEMINAR);
        return fd;

    }

    @Transactional
    private void setSeminarDocument(FileDescription fileDescription, FinalSeminar seminar) {
        fileDescription.setExposed(true);
        seminar.setDocument(fileDescription);
        seminar.setDocumentUploadDate(new Date());

        seminar = finalSeminarService.save(seminar);
        plagiarismControl.submit(fileDescription, seminar.getProject().getHeadSupervisor());

	    eventBus.post(new FinalSeminarThesisUploadedEvent(seminar));
    }

    private void assertParticipant(final User uploader, final FinalSeminar seminar) {
        final Project p = projectService.findOne(seminar.getProject().getId());
        boolean isParticipant = false;
        for (User s : p.getProjectParticipants()) {
            if (s.equals(uploader)) {
                isParticipant = true;
                break;
            }
        }
        if (!isParticipant) {
            throw new IllegalStateException("Users not a participant of the seminar project cannot upload the seminar document");
        }
    }

    private void assertMimeType(final String type) {
        final boolean requirePdf = finalSeminarSettingsService.getInstance().getThesisMustBePDF();
        if (type == null || (requirePdf && !type.equals("application/pdf"))) {
            LOGGER.debug("Faulty file type uploaded. Expected " + (requirePdf ? "\"application/pdf\"" : "non null") + ", was: \"" + type + "\"");
            throw new InvalidFileFormatException("The file must be " + (requirePdf ? "a PDF" : "non null"));
        }
    }

}
