package se.su.dsv.scipro.match;

import com.google.common.primitives.Ints;
import com.querydsl.core.types.EntityPath;
import com.querydsl.core.types.Projections;
import com.querydsl.core.types.dsl.Expressions;
import com.querydsl.core.types.dsl.StringPath;
import com.querydsl.jpa.JPAExpressions;
import com.querydsl.jpa.JPQLQuery;
import com.querydsl.jpa.impl.JPAQuery;
import org.springframework.data.domain.Pageable;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.List;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import static com.querydsl.core.types.dsl.Expressions.anyOf;

class MatchFollowUpServiceImpl implements MatchFollowUpService {

    private final Provider<EntityManager> entityManagerProvider;

    @Inject
    MatchFollowUpServiceImpl(final Provider<EntityManager> entityManagerProvider) {
        this.entityManagerProvider = entityManagerProvider;
    }

    private <T> JPAQuery<T> from(final EntityPath<T> entityPath) {
        return new JPAQuery<T>(entityManagerProvider.get()).from(entityPath);
    }

    private JPQLQuery<Long> getQuery(ApplicationPeriod applicationPeriod) {
        return JPAExpressions
                .select(QIdea.idea.count())
                .from(QIdea.idea)
                .where(QIdea.idea.published.eq(true),
                        QIdea.idea.projectType.id.eq(QTarget.target1.projectType.id))
                .where(QIdea.idea.match.supervisor.id.eq(QTarget.target1.user.id),
                        anyOf(QIdea.idea.applicationPeriod.id.eq(applicationPeriod.getId()),
                              QIdea.idea.cachedStatus.in(Idea.Status.UNMATCHED)
                        )
                );
    }

    private static Predicate<TargetResult> unmet() {
        return input -> input != null && (input.getTarget().getTarget() > input.getSubmittedAndTaken());
    }

    @Override
    public long countUnmetTargets(final ApplicationPeriod applicationPeriod) {
        List<TargetResult> list = from(QTarget.target1)
                .where(QTarget.target1.applicationPeriod.eq(applicationPeriod),
                        QTarget.target1.target.gt(0))
                .select(Projections.constructor(TargetResult.class,
                        QTarget.target1,
                        getQuery(applicationPeriod)
                ))
                .fetch();


        return list.stream()
                .filter(unmet())
                .count();
    }

    @Override
    public Iterable<TargetResult> findUnmetTargets(final ApplicationPeriod applicationPeriod, final Pageable pageable) {
        StringPath sort = Expressions.stringPath(QTarget.target1, pageable.getSort().getProperty());
        List<TargetResult> targetResults = from(QTarget.target1)
                .where(QTarget.target1.applicationPeriod.eq(applicationPeriod),
                        QTarget.target1.target.gt(0))
                .orderBy(pageable.getSort().isAscending() ? sort.asc() : sort.desc())
                .select(Projections.constructor(TargetResult.class,
                        QTarget.target1,
                        getQuery(applicationPeriod)
                ))
                .fetch();

        return targetResults.stream()
                .filter(unmet())
                .skip(Ints.checkedCast(pageable.getOffset()))
                .limit(Ints.checkedCast(pageable.getLimit()))
                .collect(Collectors.toList());
    }

    @Override
    public Set<User> findSupervisorsWithUnmetTargets(ApplicationPeriod applicationPeriod) {
        List<TargetResult> targetResults = from(QTarget.target1)
                .where(QTarget.target1.applicationPeriod.eq(applicationPeriod),
                        QTarget.target1.target.gt(0))
                .select(Projections.constructor(TargetResult.class,
                        QTarget.target1,
                        getQuery(applicationPeriod)
                ))
                .fetch();

        return targetResults.stream()
                .filter(unmet())
                .map(unmet -> unmet.getTarget().getUser())
                .collect(Collectors.toSet());
    }
}
