package se.su.dsv.scipro.plagiarism.urkund;

import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.sukat.Sukat;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.system.Username;

import javax.inject.Inject;
import java.time.Instant;
import java.time.ZoneOffset;
import java.util.Optional;

class UrkundServiceImpl implements UrkundService {
    private final UrkundApi urkundApi;
    private final UrkundSubmissionRepository urkundSubmissionRepository;
    private final Sukat sukat;

    @Inject
    UrkundServiceImpl(
            final UrkundApi urkundApi,
            final UrkundSubmissionRepository urkundSubmissionRepository,
            final Sukat sukat)
    {
        this.urkundApi = urkundApi;
        this.urkundSubmissionRepository = urkundSubmissionRepository;
        this.sukat = sukat;
    }

    @Override
    public UrkundSubmission submit(FileDescription file, User receiver) {
        Optional<UrkundSubmission> submission = getSubmission(file);
        return submission.orElseGet(() -> submitToUrkund(file, receiver));
    }

    private UrkundSubmission submitToUrkund(final FileDescription file, final User receiver) {
        return receiver.getUsernames()
                .stream()
                .filter(UrkundServiceImpl::isSuUsername)
                .findAny()
                .flatMap(sukat::email)
                .flatMap(urkundApi::analysisAddress)
                .map(analysisAddress -> {
                    UrkundResponse urkundResponse = urkundApi.submitDocument(analysisAddress, file, file.getUploader().getEmailAddress(), "", "", false);
                    return toSubmission(file, receiver, urkundResponse, analysisAddress);
                })
                .orElseGet(() -> missingAnalysisAddress(file, receiver));
    }

    private UrkundSubmission missingAnalysisAddress(final FileDescription file, final User receiver) {
        final UrkundSubmission submission = new UrkundSubmission();
        submission.setMessage(receiver.getFullName() + " has no analysis address for URKUND");
        submission.setReceiver(receiver);
        submission.setFileDescription(file);
        submission.setSubmitted(Instant.now());
        submission.setNextPoll(Instant.now());
        submission.setState(UrkundSubmission.State.SUBMISSION_FAILED);
        return urkundSubmissionRepository.save(submission);
    }

    private static boolean isSuUsername(final Username username) {
        return username.getUsername().toLowerCase().endsWith("@su.se");
    }

    @Override
    public Optional<UrkundSubmission> getSubmission(FileDescription file) {
        return urkundSubmissionRepository.findByFile(file);
    }

    private UrkundSubmission toSubmission(final FileDescription file, final User receiver, final UrkundResponse urkundResponse, String analysisAddress) {
        final UrkundSubmission submission = new UrkundSubmission();
        submission.setFileDescription(file);
        submission.setReceiver(receiver);
        submission.setAnalysisAddress(analysisAddress);
        submission.setState(urkundResponse.fold(
                error -> UrkundSubmission.State.SUBMISSION_FAILED,
                submissionInfo -> Util.getState(submissionInfo.getStatus().getState())));
        submission.setMessage(urkundResponse.getUserFriendlyMessage());
        submission.setSubmitted(urkundResponse.fold(
                e -> Instant.now(),
                submissionInfo -> submissionInfo.getSubmitted().toInstant(ZoneOffset.UTC)));
        submission.setNextPoll(submission.getSubmitted().plus(submission.getPollingDelay().getDuration()));
        return urkundSubmissionRepository.save(submission);
    }

}
