package se.su.dsv.scipro.reviewing;

import org.joda.time.Instant;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.util.Option;

import javax.persistence.*;
import java.util.Date;

@Entity
@Table(name = "Decision")
@SuppressWarnings("PMD.UnusedPrivateField")
public class Decision {
    @Id
    @GeneratedValue
    private Long id;

    @Enumerated(EnumType.STRING)
    private Status status = Status.UNDECIDED;

    @OneToOne(optional = false)
    private FileDescription thesis;

    @Basic
    private String reason;

    @Basic
    private String comment;

    @OneToOne(optional = true)
    private FileDescription attachment;

    @OneToOne(optional = true)
    private FileDescription gradingReport;

    @Temporal(TemporalType.TIMESTAMP)
    private Date requested;

    @Temporal(TemporalType.TIMESTAMP)
    private Date deadline;

    @Temporal(TemporalType.TIMESTAMP)
    private Date decisionDate;

    @ManyToOne(optional = false)
    private ReviewerApproval reviewerApproval;

    protected Decision() {} // JPA

    Decision(ReviewerApproval reviewerApproval, final FileDescription thesis, final String comment, final Date deadline) {
        this(reviewerApproval, thesis, Instant.now(), comment, deadline);
    }

    Decision(ReviewerApproval reviewerApproval, final FileDescription thesis, final Instant requested, final String comment, final Date deadline) {
        this.reviewerApproval = reviewerApproval;
        this.thesis = thesis;
        this.requested = requested.toDate();
        this.deadline = deadline;
        this.comment = comment;
    }

    public FileDescription getThesis() {
        return thesis;
    }

    public ReviewerApproval getReviewerApproval() {
        return reviewerApproval;
    }

    public Status getStatus() {
        return status;
    }

    public String getReason() {
        return reason;
    }

    public String getComment() {
        return comment;
    }

    public Option<FileDescription> getAttachment() {
        return Option.of(attachment);
    }

    public Option<FileDescription> getGradingReport() {
        return Option.of(gradingReport);
    }

    public Date getRequested() {
        return requested;
    }

    public Date getDecisionDate() {
        return decisionDate;
    }

    public Date getDeadline() {
        return deadline;
    }

    public void setDeadline(final Date deadline) {
        this.deadline = deadline;
    }

    void approve(final String reason, final Option<FileDescription> attachment, final Option<FileDescription> gradingReport) {
        decide(Status.APPROVED, reason, attachment, gradingReport);
    }

    void reject(final String reason, final Option<FileDescription> attachment, final Option<FileDescription> gradingReport) {
        decide(Status.REJECTED, reason, attachment, gradingReport);
    }

    boolean isLate() {
        return decisionDate().after(deadline);
    }

    private Date decisionDate() {
        return decisionDate == null ? new Date() : decisionDate;
    }

    private void decide(Status status, final String reason, final Option<FileDescription> attachment, final Option<FileDescription> gradingReport) {
        decideWithDecisionDate(status, reason, attachment, gradingReport, Instant.now());
    }

    private void decideWithDecisionDate(final Status status, final String reason, final Option<FileDescription> attachment, final Option<FileDescription> gradingReport, final Instant decisionDate) {
        if (this.status != Status.UNDECIDED) {
            throw new IllegalStateException();
        }
        this.decisionDate = decisionDate.toDate();
        this.status = status;
        this.reason = reason;
        this.attachment = attachment.getOrElse(() -> null);
        this.gradingReport = gradingReport.getOrElse(() -> null);
    }
}
