package se.su.dsv.scipro.reviewing;

import com.google.common.eventbus.EventBus;
import com.google.inject.persist.Transactional;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.misc.DaysService;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.util.Either;
import se.su.dsv.scipro.util.Option;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Date;

class FinalSeminarApprovalServiceImpl extends AbstractReviewerApprovalServiceImpl<FinalSeminarApproval> implements FinalSeminarApprovalService {

    private final DaysService daysService;
    private final ReviewerDeadlineSettingsService reviewerDeadlineSettingsService;

    @Inject
    FinalSeminarApprovalServiceImpl(
            final Provider<EntityManager> em,
            final FileDescriptionService fileDescriptionService,
            final EventBus eventBus,
            final DaysService daysService,
            final ReviewerDeadlineSettingsService reviewerDeadlineSettingsService)
    {
        super(em, FinalSeminarApproval.class, QFinalSeminarApproval.finalSeminarApproval, eventBus, fileDescriptionService);
        this.daysService = daysService;
        this.reviewerDeadlineSettingsService = reviewerDeadlineSettingsService;
    }

    @Override
    @Transactional
    public Either<AlreadyRequested, FinalSeminarApproval> requestApproval(final Project project, FileUpload thesis, final String comment) {
        final FileDescription storedThesis = fileDescriptionService.storeFile(thesis);

        final Either<AlreadyRequested, FinalSeminarApproval> finalSeminarApproval = findBy(project)
                .map(earlier -> addNewRequest(comment, storedThesis, earlier))
                .getOrElse(() -> Either.right(newApproval(project, storedThesis, comment)));

        final Either<AlreadyRequested, FinalSeminarApproval> persisted =
                finalSeminarApproval.map(this::save);
        persisted.foreach(fsa -> eventBus.post(new FinalSeminarApprovalRequestEvent(fsa)));
        return persisted;
    }

    private FinalSeminarApproval newApproval(Project project, FileDescription storedThesis, String comment) {
        final Date deadline = getDeadline(new Date());
        return new FinalSeminarApproval(project, storedThesis, comment, deadline);
    }

    @Override
    protected Date getDeadline(final Date requested) {
        final ReviewerDeadlineSettings reviewerDeadlineSettings = reviewerDeadlineSettingsService.getInstance();
        return daysService.workDaysAfter(requested, reviewerDeadlineSettings.getFinalSeminarApproval());
    }

    @Override
    public Option<FinalSeminarApproval> findBy(final Project project) {
        return Option.of(findOne(QFinalSeminarApproval.finalSeminarApproval.project.eq(project)));
    }

    @Override
    public FinalSeminarApproval findByProject(Project project) {
        return findOne(QFinalSeminarApproval.finalSeminarApproval.project.eq(project));
    }
}
