package se.su.dsv.scipro.system;

import com.google.inject.persist.Transactional;
import com.querydsl.core.types.Predicate;
import com.querydsl.core.types.dsl.EntityPathBase;
import com.querydsl.core.types.dsl.Expressions;
import com.querydsl.core.types.dsl.StringPath;
import com.querydsl.jpa.impl.JPAQuery;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.data.querydsl.QueryDslPredicateExecutor;

import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.io.Serializable;
import java.util.List;

public class GenericRepo<T, ID extends Serializable> implements QueryDslPredicateExecutor<T>, JpaRepository<T, ID> {

    private final Provider<EntityManager> em;
    private final Class<T> clazz;
    private final EntityPathBase<T> path;

    public GenericRepo(Provider<EntityManager> em, Class<T> clazz, EntityPathBase<T> path) {
        this.em = em;
        this.clazz = clazz;
        this.path = path;
    }

    protected EntityManager em() {
        return em.get();
    }

    protected JPAQuery<T> createQuery() {
        return new JPAQuery<T>(em.get()).from(path);
    }

    private JPAQuery<T> applyPageable(JPAQuery<T> query, Pageable pageable) {
        JPAQuery<T> pagedQuery = paged(query, pageable.getOffset(), pageable.getLimit());
        return sorted(pagedQuery, pageable.getSort());
    }

    private JPAQuery<T> paged(JPAQuery<T> query, long offset, long limit) {
        return query.limit(limit).offset(offset);
    }

    private JPAQuery<T> sorted(JPAQuery<T> query, Sort sort) {
        if (sort == null) {
            return query;
        }
        else {
            StringPath property = Expressions.stringPath(path, sort.getProperty());
            return query.orderBy(sort.isAscending() ? property.asc() : property.desc());
        }
    }

    @Override
    public List<T> findAll() {
        return createQuery().fetch();
    }

    @Override
    public long count() {
        return createQuery().fetchCount();
    }

    @Override
    @Transactional
    public void delete(ID id) {
        em.get().remove(findOne(id));
    }

    @Override
    @Transactional
    public void delete(T entity) {
        em.get().remove(em.get().contains(entity) ? entity : em.get().merge(entity));
    }

    @Override
    @Transactional
    public <S extends T> S save(S entity) {
        EntityManager entityManager = em.get();
        if (entityManager.contains(entity)) {
            return entityManager.merge(entity);
        } else {
            entityManager.persist(entity);
            return entity;
        }
    }

    @Override
    public T findOne(ID id) {
        return em.get().find(clazz, id);
    }

    @Override
    public T findOne(Predicate predicate) {
        return createQuery().where(predicate).fetchOne();
    }

    @Override
    public List<T> findAll(Predicate predicate) {
        return createQuery().where(predicate).fetch();
    }

    @Override
    public List<T> findAll(Predicate predicate, Pageable pageable) {
        return applyPageable(createQuery(), pageable).where(predicate).fetch();
    }

    @Override
    public long count(Predicate predicate) {
        return createQuery().where(predicate).fetchCount();
    }
}
