package se.su.dsv.scipro.util;

import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

public abstract class Option<A> {
    public static <A> Option<A> empty() {
        return new None<>();
    }
    public static <A> Option<A> of(A a) {
        return a == null ? empty() : new Some<>(a);
    }

    public abstract <X> X fold(Supplier<X> ifEmpty, Function<? super A, X> f);

    public <B> Option<B> flatMap(Function<? super A, Option<B>> f) {
        return fold(Option::empty, f);
    }

    public <B> Option<B> map(Function<? super A, B> f) {
        return flatMap(f.andThen(Option::of));
    }

    public Option<A> filter(Predicate<? super A> p) {
        return flatMap(a -> p.test(a) ? this : empty());
    }

    public Option<A> orElse(Supplier<Option<A>> alternative) {
        return fold(alternative, Option::of);
    }

    public A getOrElse(Supplier<A> alternative) {
        return fold(alternative, a -> a);
    }

    public void foreach(Consumer<? super A> sideEffect) {
        fold(() -> null, a -> { sideEffect.accept(a); return null; });
    }

    public boolean forall(Predicate<? super A> p) {
        return fold(() -> true, p::test);
    }

    public boolean exists(Predicate<? super A> p) {
        return fold(() -> false, p::test);
    }

    public boolean isPresent() {
        return fold(() -> false, a -> true);
    }

    @Override
    public String toString() {
        return fold(() -> "None", a -> "Some(" + a.toString() + ")");
    }

    private static final class None<A> extends Option<A> {
        @Override
        public <X> X fold(final Supplier<X> ifEmpty, final Function<? super A, X> f) {
            return ifEmpty.get();
        }

        @Override
        public boolean equals(final Object obj) {
            return obj instanceof None;
        }

        @Override
        public int hashCode() {
            return 0;
        }
    }

    private static final class Some<A> extends Option<A> {
        private final A a;

        public Some(final A a) {
            super();
            this.a = a;
        }

        @Override
        public <X> X fold(final Supplier<X> ifEmpty, final Function<? super A, X> f) {
            return f.apply(a);
        }

        @Override
        public boolean equals(final Object obj) {
            if (this == obj) return true;
            if (!(obj instanceof Some)) return false;
            final Some<?> some = (Some<?>) obj;
            return Objects.equals(a, some.a);
        }

        @Override
        public int hashCode() {
            return Objects.hash(a);
        }
    }
}
