package se.su.dsv.scipro.integration.daisy.workers;

import org.joda.time.LocalDateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.daisyExternal.http.DaisyAPI;
import se.su.dsv.scipro.finalseminar.FinalSeminar;
import se.su.dsv.scipro.finalseminar.FinalSeminarParticipation;
import se.su.dsv.scipro.finalseminar.FinalSeminarService;
import se.su.dsv.scipro.finalthesis.FinalThesis;
import se.su.dsv.scipro.finalthesis.FinalThesisService;
import se.su.dsv.scipro.io.ExternalExporter;
import se.su.dsv.scipro.io.dto.*;
import se.su.dsv.scipro.io.exceptions.ExternalExportException;
import se.su.dsv.scipro.io.facade.ExporterFacade;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectRepo;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.QProject;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.Option;
import se.su.dsv.scipro.workerthreads.AbstractWorker;

import javax.inject.Inject;
import javax.inject.Named;
import java.util.Calendar;
import java.util.Date;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import static com.querydsl.core.types.dsl.Expressions.anyOf;

@Named
public class ProjectExporter extends AbstractWorker {

    private static final Logger LOG = LoggerFactory.getLogger(ProjectExporter.class);
    private static final int MAX_TITLE_LENGTH = 255;

    private final ProjectRepo projectRepo;
    private final ExporterFacade exporterFacade;
    private final DaisyAPI daisyAPI;
    private final ExternalExporter externalExporter;
    private final FinalSeminarService finalSeminarService;
    private final FinalThesisService finalThesisService;

    @Inject
    public ProjectExporter(ProjectRepo projectRepo,
                           ExporterFacade exporterFacade,
                           DaisyAPI daisyAPI,
                           ExternalExporter externalExporter,
                           final FinalSeminarService finalSeminarService,
                           final FinalThesisService finalThesisService)
    {
        this.projectRepo = projectRepo;
        this.exporterFacade = exporterFacade;
        this.daisyAPI = daisyAPI;
        this.externalExporter = externalExporter;
        this.finalSeminarService = finalSeminarService;
        this.finalThesisService = finalThesisService;
    }

    @Override
    protected void doWork() {
        beginTransaction();
        Iterable<Project> projects = projectRepo.findAll(
                QProject.project.projectType.degreeLevel.in(DegreeLevel.FIRST_CYCLE, DegreeLevel.SECOND_CYCLE)
                        .and(QProject.project.headSupervisor.unit.isNotNull())
                        .and(anyOf(
                                QProject.project.lastModified.after(oneWeekAgo()),
                                QProject.project.projectStatus.eq(ProjectStatus.ACTIVE)
                        ))
                        .and(QProject.project.headSupervisor.unit.identifier.isNotNull()));

        for (Project project : projects) {
            try {
                exportProject(project);
            }
            catch (Exception e) {
                LOG.info("Failed to export project ({}) to Daisy", project, e);
            }
        }
        commitTransaction();
    }

    private static Date oneWeekAgo() {
        return new LocalDateTime().minusWeeks(1).toDate();
    }

    private void exportProject(Project project) throws ExternalExportException {
        if (project.getIdentifier() == null) {
            createProject(project);
        } else {
            final Option<FinalThesis> finalThesis = Option.of(finalThesisService.findByProject(project));
            updateProjectData(project, finalThesis);
            updateAuthors(project);
            updateHeadSupervisor(project);
            updateReviewers(project);
            updateCoSupervisors(project);

            FinalSeminar finalSeminar = finalSeminarService.findByProject(project);
            if (finalSeminar != null) {
                updateParticipants(project.getIdentifier(), finalSeminar.getActiveParticipations()
                        .stream()
                        .filter(isApproved())
                        .map(toUser())
                        .filter(onlyDaisyUsers()));
                updateOpponents(project.getIdentifier(), finalSeminar.getOppositions()
                        .stream()
                        .filter(isApproved())
                        .map(toUser())
                        .filter(onlyDaisyUsers()));
            }
        }
    }

    private static Predicate<User> onlyDaisyUsers() {
        return input -> input.getIdentifier() != null;
    }

    private static Function<FinalSeminarParticipation, User> toUser() {
        return FinalSeminarParticipation::getUser;
    }

    private static Predicate<FinalSeminarParticipation> isApproved() {
        return FinalSeminarParticipation::isApproved;
    }

    private void updateOpponents(Long projectId, final Stream<User> opponents) {
        updateParticipations(projectId, opponents, Role.OPPONENT);
    }

    private void updateParticipants(final Long projectId, final Stream<User> participants) {
        updateParticipations(projectId, participants, Role.ACTIVE_PARTICIPATION);
    }

    private void updateParticipations(final Long projectId, final Stream<User> users, final Role role) {
        final Set<ProjectParticipant> participants = daisyAPI.getContributors(projectId);
        users.forEach(user -> {
            if (!isParticipant(user, role, participants)) {
                Person person = new Person();
                person.setId(user.getIdentifier());
                ProjectParticipant participant = new ProjectParticipant();
                participant.setRole(role);
                participant.setPerson(person);
                daisyAPI.addContributor(projectId, participant);
            }
        });
    }

    private boolean isParticipant(final User user, final Role role, final Set<ProjectParticipant> contributors) {
        return contributors.stream().anyMatch(contributor -> contributor.getRole() == role && contributor.getPerson().getId().equals(user.getIdentifier()));
    }

    private void createProject(Project project) throws ExternalExportException {
        exporterFacade.exportProject(project, project.getHeadSupervisor().getUnit());
    }

    private void updateProjectData(Project project, final Option<FinalThesis> finalThesis) {
        Option<Thesis> daisyThesis = daisyAPI.getThesis(project.getIdentifier());

        if (daisyThesis.isPresent()) {
            daisyThesis.foreach(thesis -> updateInDaisy(project, finalThesis, thesis));
        } else {
            LOG.info("Daisy project missing for project: {}", project);
        }
    }

    private void updateInDaisy(final Project project, final Option<FinalThesis> finalThesis, final Thesis daisyThesis) {
        Unit unit = new Unit();
        unit.setId(project.getHeadSupervisor().getUnit().getIdentifier());

        Calendar startDate = Calendar.getInstance();
        startDate.setTime(project.getDateCreated());

        String title = finalThesis
                .map(FinalThesis::getSwedishTitle)
                .orElse(() -> finalThesis.map(FinalThesis::getEnglishTitle))
                .getOrElse(project::getTitle);
        String englishTitle = finalThesis
                .map(FinalThesis::getEnglishTitle)
                .getOrElse(daisyThesis::getTitleEn);

        Thesis thesis = new Thesis();
        thesis.setId(project.getIdentifier());
        thesis.setLevel(mapLevel(project));
        thesis.setTitle(truncate(title));
        thesis.setTitleEn(englishTitle != null ? truncate(englishTitle) : null);
        thesis.setUnit(unit);
        thesis.setAborted(project.getProjectStatus() == ProjectStatus.INACTIVE);
        thesis.setFinished(daisyThesis.isFinished());
        thesis.setStartDate(startDate);

        daisyAPI.updateThesis(thesis);
    }

    private static String truncate(String projectTitle) {
        return projectTitle.length() > MAX_TITLE_LENGTH ? projectTitle.substring(0, MAX_TITLE_LENGTH) : projectTitle;
    }

    private static Level mapLevel(Project project) {
        return Level.valueOf(project.getProjectType().getDegreeLevel().name());
    }

    private void updateAuthors(Project project) throws ExternalExportException {
        Set<StudentProjectParticipant> authors = daisyAPI.getAuthors(project.getIdentifier());
        for (StudentProjectParticipant author : authors) {
            if (authorIsRemoved(author.getPerson().getId(), project.getProjectParticipants())) {
                daisyAPI.deleteAuthor(project.getIdentifier(), author.getPerson().getId());
            }
        }
        for (User user : project.getProjectParticipants()) {
            if (authorIsMissing(user.getIdentifier(), authors)) {
                externalExporter.addStudentToProject(project, user);
            }
        }
    }

    private boolean authorIsMissing(Long identifier, Iterable<StudentProjectParticipant> authors) {
        for (StudentProjectParticipant author : authors) {
            if (author.getPerson().getId().equals(identifier)) {
                return false;
            }
        }
        return true;
    }

    private boolean authorIsRemoved(Long id, Iterable<User> authors) {
        for (User author : authors) {
            if (id.equals(author.getIdentifier())) {
                return false;
            }
        }
        return true;
    }

    private void updateHeadSupervisor(Project project) throws ExternalExportException {
        boolean add = true;
        Set<ProjectParticipant> contributors = daisyAPI.getContributors(project.getIdentifier());
        for (ProjectParticipant contributor : contributors) {
            if (contributor.getRole() == Role.SUPERVISOR) {
                if (contributor.getPerson().getId().equals(project.getHeadSupervisor().getIdentifier())) {
                    add = false;
                } else {
                    daisyAPI.deleteContributor(project.getIdentifier(), contributor.getPerson().getId());
                }
            }
        }
        if (add) {
            externalExporter.addContributorToProject(project, project.getHeadSupervisor(), Role.SUPERVISOR);
        }
    }

    private void updateReviewers(Project project) throws ExternalExportException {
        Set<ProjectParticipant> contributors = daisyAPI.getContributors(project.getIdentifier());
        for (ProjectParticipant contributor : contributors) {
            if (contributor.getRole() == Role.EXAMINER &&
                    reviewerIsRemoved(contributor.getPerson().getId(), project.getReviewers())) {
                daisyAPI.deleteContributor(project.getIdentifier(), contributor.getPerson().getId());
            }
        }
        for (User reviewer : project.getReviewers()) {
            if (reviewerIsMissing(reviewer, contributors)) {
                externalExporter.addContributorToProject(project, reviewer, Role.EXAMINER);
            }
        }
    }

    private boolean reviewerIsRemoved(Long id, Iterable<User> projectReviewers) {
        for (User reviewer : projectReviewers) {
            if (id.equals(reviewer.getIdentifier())) {
                return false;
            }
        }
        return true;
    }

    private boolean reviewerIsMissing(User reviewer, Iterable<ProjectParticipant> contributors) {
        for (ProjectParticipant contributor : contributors) {
            if (contributor.getRole() == Role.EXAMINER && contributor.getPerson().getId().equals(reviewer.getIdentifier())) {
                return false;
            }
        }
        return reviewer.getIdentifier() != null;
    }

    private void updateCoSupervisors(Project project) throws ExternalExportException {
        Set<ProjectParticipant> contributors = daisyAPI.getContributors(project.getIdentifier());
        for (ProjectParticipant contributor : contributors) {
            if (contributor.getRole() == Role.ASSISTANT_SUPERVISOR &&
                    coSupervisorIsRemoved(contributor.getPerson().getId(), project.getCoSupervisors())) {
                daisyAPI.deleteContributor(project.getIdentifier(), contributor.getPerson().getId());
            }
        }
        for (User coSupervisor : project.getCoSupervisors()) {
            if (coSupervisorIsMissing(coSupervisor, contributors)) {
                externalExporter.addContributorToProject(project, coSupervisor, Role.ASSISTANT_SUPERVISOR);
            }
        }
    }

    private boolean coSupervisorIsRemoved(Long id, Set<User> coSupervisors) {
        for (User coSupervisor : coSupervisors) {
            if (id.equals(coSupervisor.getIdentifier())) {
                return false;
            }
        }
        return true;
    }

    private boolean coSupervisorIsMissing(User coSupervisor, Set<ProjectParticipant> contributors) {
        for (ProjectParticipant contributor : contributors) {
            if (contributor.getRole() == Role.ASSISTANT_SUPERVISOR && contributor.getPerson().getId().equals(coSupervisor.getIdentifier())) {
                return false;
            }
        }
        return coSupervisor.getIdentifier() != null;
    }
}
