package se.su.dsv.scipro.datatables.project;

import com.google.common.eventbus.EventBus;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.sort.SortOrder;
import org.apache.wicket.extensions.markup.html.repeater.data.table.AbstractColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.link.AbstractLink;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import scala.runtime.AbstractFunction1;
import se.su.dsv.scipro.admin.pages.AdminEditProjectPage;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.components.datatables.MultipleUsersColumn;
import se.su.dsv.scipro.components.datatables.UserColumn;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.dataproviders.FilteredDataProvider;
import se.su.dsv.scipro.datatables.AjaxCheckboxWrapper;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.notifications.dataobject.ProjectEvent;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.ReviewerAssignedEvent;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.*;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.*;

public class ProjectDataPanel extends Panel {

    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private ProjectService projectService;
    @Inject
    private EventBus eventBus;
    @Inject
    private NotificationController notificationController;
    @Inject
    private UserService userService;

    private ExportableDataPanel<Project, String> exportableDataPanel;
    private ProjectService.Filter filter;
    private FencedFeedbackPanel feedback;

    public ProjectDataPanel(String id) {
        super(id);
        filter = new ProjectService.Filter();
        filter.setFilterExternal(false);
        final List<ProjectType> defaultProjectTypes
                = projectTypeService.findByDegreeLevels(Arrays.asList(DegreeLevel.FIRST_CYCLE, DegreeLevel.SECOND_CYCLE));
        filter.setProjectTypes(defaultProjectTypes);
        filter.setStatuses(new ArrayList<>(Collections.singletonList(ProjectStatus.ACTIVE)));
        add(new ProjectFilterPanel("filterForm", Model.of(filter)) {
            @Override
            protected void filterUpdated(final AjaxRequestTarget target) {
                target.add(exportableDataPanel);
            }
        });
        addCreateNewProjectLink();
        addFeedbackPanel();
        addDataTable();
    }

    private void addDataTable() {
        SortableDataProvider<Project, String> provider = new FilteredDataProvider<>(projectService, Model.of(filter));
        provider.setSort("dateCreated", SortOrder.DESCENDING);
        exportableDataPanel = new ExportableDataPanel<>("dataTable", createColumns(), provider);
        add(exportableDataPanel);
    }

    private List<IColumn<Project, String>> createColumns() {
        List<IColumn<Project, String>> columns = new ArrayList<>();
        final boolean isAdmin = SciProSession.get().authorizedForRole(Roles.ADMIN);
        columns.add(new DateColumn<Project>(Model.of("Started"), "dateCreated", "dateCreated"));
        if (isAdmin) {
            columns.add(titleColumn());
        }
        else {
            columns.add(new PropertyColumn<>(Model.of("Title"), "title", "title"));
        }
        columns.add(new PropertyColumn<Project, String>(Model.of("Type"), "projectType.name", "projectType.name"));
        columns.add(authorsColumn());
        columns.add(new PropertyColumn<Project, String>(Model.of("Status"), "projectStatus", "projectStatus"));
        columns.add(supervisorColumn());
        if (isAdmin) {
            columns.add(reviewerColumn());
        }
        else {
            columns.add(new UserColumn<Project, String>(Model.of("Reviewer")) {
                @Override
                protected IModel<User> getUser(final IModel<Project> rowModel) {
                    return new PropertyModel<>(rowModel, "reviewer");
                }
            });
        }
        columns.add(new PropertyColumn<Project, String>(Model.of("Research area"), "headSupervisor.researchAreaNames"));
        if (isAdmin) {
            columns.add(inactivateCheckbox());
        }
        return columns;
    }

    private AbstractColumn<Project, String> supervisorColumn() {
        return new PropertyColumn<Project, String>(Model.of("Head Superv."), "headSupervisor.fullName") {
            @Override
            public void populateItem(Item<ICellPopulator<Project>> cellItem, String componentId, IModel<Project> rowModel) {
                User supervisor = rowModel.getObject().getHeadSupervisor();
                if (supervisor != null) {
                    cellItem.add(new UserLinkPanel(componentId, new PropertyModel<User>(rowModel, "headSupervisor")));
                } else {
                    cellItem.add(new Label(componentId, "-"));
                }
            }
        };
    }

    private AbstractColumn<Project, String> authorsColumn() {
        return new MultipleUsersColumn<Project, String>(Model.of("Authors")) {
            @Override
            public IModel<? extends List<User>> getUsers(final IModel<Project> rowModel) {
                return new ListAdapterModel<>(new PropertyModel<Set<User>>(rowModel, "projectParticipants"));
            }
        };
    }

    private AbstractColumn<Project, String> titleColumn() {
        return new PropertyColumn<Project, String>(Model.of("Title"), "title") {
            @Override
            public void populateItem(Item<ICellPopulator<Project>> components, String s, final IModel<Project> model) {
                components.add(LinkWrapper.apply(s, new AbstractFunction1<String, AbstractLink>() {
                    @Override
                    public AbstractLink apply(final String id) {
                        final PageParameters pp = new PageParameters();
                        pp.set(PageParameterKeys.MAP.get(Project.class), model.getObject().getId());
                        final BookmarkablePageLink<Void> link = new BookmarkablePageLink<>(id, AdminEditProjectPage.class, pp);
                        link.setBody(Model.of(model.getObject().getTitle()));
                        return link;
                    }
                }));
            }
        };
    }

    private AbstractColumn<Project, String> reviewerColumn() {
        return new PropertyColumn<Project, String>(Model.of("Reviewer"), "reviewer") {
            @Override
            public void populateItem(final Item<ICellPopulator<Project>> cellItem, final String componentId, final IModel<Project> rowModel) {
                cellItem.add(new EmployeeAutoCompleteDivPanel(componentId, new DetachableServiceModel<>(userService, rowModel.getObject().getReviewer())) {
                    @Override
                    public void onNewEmployeeSelection(final AjaxRequestTarget target, final User newSelection) {
                        Project project = rowModel.getObject();
                        if (project.getHeadSupervisor().equals(newSelection)) {
                            error(getString("supervisor.reviewer.error"));
                        } else if (!newSelection.getRoles().contains(Roles.REVIEWER)) {
                            error(getString("not.reviewer.error"));
                        } else {
                            for (User currentReviewer : project.getReviewers()) {
                                project.removeReviewer(currentReviewer);
                            }
                            project.addReviewer(newSelection);
                            notifyReviewersChanged(project);
                            eventBus.post(new ReviewerAssignedEvent(project, newSelection));

                            projectService.save(project);
                            info(getString("reviewer.updated"));
                        }
                        target.add(feedback);
                        target.add(exportableDataPanel);
                    }
                });
            }
        };
    }

    private AbstractColumn<Project, String> inactivateCheckbox() {
        return new AbstractExportableColumn<Project, String>(Model.of("Active")) {
            @Override
            public void populateItem(Item<ICellPopulator<Project>> cellItem, String componentId, final IModel<Project> model) {
                final Project project = model.getObject();
                final boolean currentState = !project.getProjectStatus().equals(ProjectStatus.INACTIVE);
                cellItem.add(new AjaxCheckboxWrapper(componentId, Model.of(currentState)) {
                    @Override
                    public void onChange(AjaxRequestTarget target, boolean selected) {
                        projectService.toggleActiveStatus(model.getObject(), selected);
                        success(getString("inactive", model));
                        target.add(feedback);
                    }
                });
            }

            @Override
            public IModel<Boolean> getDataModel(IModel<Project> rowModel) {
                return Model.of(!rowModel.getObject().getProjectStatus().equals(ProjectStatus.INACTIVE));
            }
        };
    }

    private void notifyReviewersChanged(Project project) {
        notificationController.notifyProject(project, ProjectEvent.Event.REVIEWERS_CHANGED, new NotificationSource());
    }

    private void addFeedbackPanel() {
        feedback = new FencedFeedbackPanel("feedback", this);
        feedback.setOutputMarkupId(true);
        add(feedback);
    }

    private void addCreateNewProjectLink() {
        AbstractLink newLink = new BookmarkablePageLink<Void>("newLink", AdminEditProjectPage.class) {
            @Override
            protected void onConfigure() {
                super.onConfigure();
                setVisibilityAllowed(SciProSession.get().authorizedForRole(Roles.ADMIN));
            }
        };
        add(newLink);
    }
}