package se.su.dsv.scipro.finalseminar;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.AjaxLink;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.upload.FileUpload;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.StringResourceModel;
import se.su.dsv.scipro.components.AbstractUploadForm;
import se.su.dsv.scipro.components.DateLabel;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.files.WicketFileUpload;
import se.su.dsv.scipro.repository.panels.ViewAttachmentPanel;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.util.AjaxConfirmationLink;

import javax.inject.Inject;
import java.util.Date;

public class SeminarThesisPanel extends GenericPanel<FinalSeminar> {
    static final String DEADLINE_INFORMATION = "deadlineInformation";
    static final String DATE = "date";
    static final String UPLOAD = "upload";
    static final String NO_THESIS = "noThesis";
    static final String DOWNLOAD_THESIS = "download";
    static final String DELETE_THESIS = "delete";
    public static final String FEEDBACK = "feedback";

    @Inject
    private FinalSeminarService seminarService;

    public SeminarThesisPanel(final String id, final IModel<FinalSeminar> model) {
        super(id, model);

        add(new FencedFeedbackPanel(FEEDBACK, this));

        IModel<Date> deadline = new LoadableDetachableModel<Date>() {
            @Override
            protected Date load() {
                return seminarService.thesisUploadDeadline(model.getObject());
            }
        };

        setOutputMarkupPlaceholderTag(true);

        addDeadlineInformation(deadline);
        add(new NewSeminarDocumentUploadForm("upload", getModel()));
        addDownloadLink();
        addDeleteLink();
        addNoThesisText();
    }

    protected void onDelete() {
    }

    private void deleteThesis(AjaxRequestTarget target) {
        FinalSeminar seminar = seminarService.deleteThesis(getSeminar());
        setModelObject(seminar);

        target.add(this);
        onDelete();
    }

    private FinalSeminar getSeminar() {
        return getModelObject();
    }

    private void addNoThesisText() {
        add(new Label(NO_THESIS, new StringResourceModel("noThesis", SeminarThesisPanel.this, null)) {
            @Override
            protected void onConfigure() {
                super.onConfigure();
                setVisibilityAllowed(!seminarService.hasThesis(getModelObject()));
            }
        });
    }

    private void addDeleteLink() {
        AjaxLink<Void> deleteLink = new AjaxConfirmationLink<Void>(DELETE_THESIS, "Are you sure you want to delete the final seminar thesis file?") {
            @Override
            public void onClick(AjaxRequestTarget target) {
                deleteThesis(target);
            }
            @Override
            protected void onConfigure() {
                super.onConfigure();
                boolean isSupervisor = getSeminar().getProject().isSupervisor(SciProSession.get().getUser());
                setVisibilityAllowed(seminarService.hasThesis(getSeminar()) && isSupervisor);
            }
        };
        add(deleteLink);
    }

    private void addDownloadLink() {
        add(new ViewAttachmentPanel(DOWNLOAD_THESIS, new PropertyModel<FileDescription>(getDefaultModel(), "document"), DateStyle.DATETIME));
    }

    private class NewSeminarDocumentUploadForm extends AbstractUploadForm<FinalSeminar> {
        @Inject
        private FinalSeminarUploadController seminarUploadController;

        private NewSeminarDocumentUploadForm(final String id, final IModel<FinalSeminar> seminar) {
            super(id, seminar);
        }

        @Override
        protected void onNewUpload(final FileUpload fileUpload) {
            if (SciProSession.exists()) {
                seminarUploadController.storeSeminarDocument(
                        WicketFileUpload.from(fileUpload), getModelObject());
            }
        }

        @Override
        protected void onFinished(AjaxRequestTarget target) {
            FinalSeminar seminar = getModelObject();
            SeminarThesisPanel.this.setModelObject(seminar);

            SeminarThesisPanel.this.success(getString("upload.success"));

            setResponsePage(ProjectFinalSeminarPage.class);
        }

        @Override
        protected void onConfigure() {
            super.onConfigure();
            setVisibilityAllowed(getSeminar().getProject().isParticipant(SciProSession.get().getUser()) && !seminarService.hasThesis(getSeminar()));
            if (!getSeminar().getProject().isFinalSeminarRuleExempted()) {
                setEnabled(!getSeminar().isDeleted() && !seminarService.hasDeadlinePassed(getSeminar()));
            }
            else {
                setEnabled(true);
            }
        }
    }

    private void addDeadlineInformation(final IModel<Date> deadline) {
        WebMarkupContainer deadlineInformation = new WebMarkupContainer(DEADLINE_INFORMATION) {
            @Override
            protected void onConfigure() {
                setVisible(getModelObject().getId() != null);
            }
        };
        deadlineInformation.add(new DateLabel(DATE, deadline, DateStyle.DATETIME));
        add(deadlineInformation);
    }
}
